"""
Custom implementation of

    Shaw et al., 2019
    MRI k-Space Motion Artefact Augmentation:
    Model Robustness and Task-Specific Uncertainty

"""

import warnings
import torch
import numpy as np
from tqdm import tqdm
import SimpleITK as sitk
from scipy.linalg import logm, expm
from ....utils import is_image_dict
from ....torchio import INTENSITY, DATA, AFFINE
from .. import Interpolation
from .. import RandomTransform


class RandomMotion(RandomTransform):
    def __init__(
            self,
            degrees=10,
            translation=10,  # in mm
            num_transforms=2,
            image_interpolation=Interpolation.LINEAR,
            proportion_to_augment=0.5,
            seed=None,
            verbose=False,
            ):
        super().__init__(seed=seed, verbose=verbose)
        self.degrees_range = self.parse_degrees(degrees)
        self.translation_range = self.parse_translation(translation)
        self.num_transforms = num_transforms
        self.image_interpolation = image_interpolation
        self.proportion_to_augment = self.parse_probability(
            proportion_to_augment,
            'proportion_to_augment',
        )

    def apply_transform(self, sample):
        for image_name, image_dict in sample.items():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] != INTENSITY:
                continue
            params = self.get_params(
                self.degrees_range,
                self.translation_range,
                self.num_transforms,
                self.proportion_to_augment
            )
            times_params, degrees_params, translation_params, do_it = params
            keys = (
                'random_motion_times',
                'random_motion_degrees',
                'random_motion_translation',
                'random_motion_do',
            )
            for key, p in zip(keys, params):
                sample[image_name][key] = p
            if not do_it:
                return sample
            if (image_dict[DATA][0] < -0.1).any():
                # I use -0.1 instead of 0 because Python was warning me when
                # a value in a voxel was -7.191084e-35
                # There must be a better way of solving this
                message = (
                    f'Image "{image_name}" from "{image_dict["stem"]}"'
                    ' has negative values.'
                    ' Results can be unexpected because the transformed sample'
                    ' is computed as the absolute values'
                    ' of an inverse Fourier transform'
                )
                warnings.warn(message)
            image = self.nib_to_sitk(
                image_dict[DATA][0],
                image_dict[AFFINE],
            )
            transforms = self.get_rigid_transforms(
                degrees_params,
                translation_params,
                image,
            )
            transforms = self.demean_transforms(
                transforms,
                times_params,
            )
            image_dict[DATA] = self.add_artifact(
                image,
                transforms,
                times_params,
                self.image_interpolation,
            )
            # Add channels dimension
            image_dict[DATA] = image_dict[DATA][np.newaxis, ...]
            image_dict[DATA] = torch.from_numpy(image_dict[DATA])
        return sample

    @staticmethod
    def get_params(
            degrees_range,
            translation_range,
            num_transforms,
            probability,
            perturbation=0.3,
            ):
        """
        If perturbation is 0, the intervals between movements are constant
        """
        degrees_params = get_params_array(
            degrees_range, num_transforms)
        translation_params = get_params_array(
            translation_range, num_transforms)
        step = 1 / (num_transforms + 1)
        times = torch.arange(0, 1, step)[1:]
        noise = torch.FloatTensor(num_transforms)
        noise.uniform_(-step * perturbation, step * perturbation)
        times += noise
        times_params = times.numpy()
        do_it = torch.rand(1) < probability
        return times_params, degrees_params, translation_params, do_it

    def get_rigid_transforms(self, degrees_params, translation_params, image):
        center_ijk = np.array(image.GetSize()) / 2
        center_lps = image.TransformContinuousIndexToPhysicalPoint(center_ijk)
        identity = np.eye(4)
        matrices = [identity]
        for degrees, translation in zip(degrees_params, translation_params):
            radians = np.radians(degrees).tolist()
            motion = sitk.Euler3DTransform()
            motion.SetCenter(center_lps)
            motion.SetRotation(*radians)
            motion.SetTranslation(translation.tolist())
            motion_matrix = self.transform_to_matrix(motion)
            matrices.append(motion_matrix)
        transforms = [self.matrix_to_transform(m) for m in matrices]
        return transforms

    def demean_transforms(self, transforms, times):
        matrices = [self.transform_to_matrix(t) for t in transforms]
        times = np.insert(times, 0, 0)
        times = np.append(times, 1)
        weights = np.diff(times)
        mean = self.matrix_average(matrices, weights=weights)
        inverse_mean = np.linalg.inv(mean)
        demeaned_matrices = [inverse_mean @ matrix for matrix in matrices]
        demeaned_transforms = [
            self.matrix_to_transform(m) for m in demeaned_matrices]
        return demeaned_transforms

    @staticmethod
    def transform_to_matrix(transform):
        matrix = np.eye(4)
        rotation = np.array(transform.GetMatrix()).reshape(3, 3)
        matrix[:3, :3] = rotation
        matrix[:3, 3] = transform.GetTranslation()
        return matrix

    @staticmethod
    def matrix_to_transform(matrix):
        transform = sitk.Euler3DTransform()
        rotation = matrix[:3, :3].flatten().tolist()
        transform.SetMatrix(rotation)
        transform.SetTranslation(matrix[:3, 3])
        return transform

    def resample_images(self, image, transforms, interpolation):
        floating = reference = image
        default_value = np.float64(sitk.GetArrayViewFromImage(image).min())
        transforms = transforms[1:]  # first is identity
        images = [image]  # first is identity
        trsfs = tqdm(transforms, leave=False) if self.verbose else transforms
        for transform in trsfs:
            resampler = sitk.ResampleImageFilter()
            resampler.SetInterpolator(interpolation.value)
            resampler.SetReferenceImage(reference)
            resampler.SetOutputPixelType(sitk.sitkFloat32)
            resampler.SetDefaultPixelValue(default_value)
            resampler.SetTransform(transform)
            resampled = resampler.Execute(floating)
            images.append(resampled)
        return images

    @staticmethod
    def sort_spectra(spectra, times):
        """
        Use original spectrum to fill the center of K-space
        """
        num_spectra = len(spectra)
        if np.any(times > 0.5):
            index = np.where(times > 0.5)[0].min()
        else:
            index = num_spectra - 1
        spectra[0], spectra[index] = spectra[index], spectra[0]

    def add_artifact(
            self,
            image,
            transforms,
            times,
            interpolation: Interpolation,
            ):
        images = self.resample_images(image, transforms, interpolation)
        arrays = [sitk.GetArrayViewFromImage(im) for im in images]
        arrays = [array.transpose() for array in arrays]  # ITK to NumPy
        arrays = tqdm(arrays, leave=False) if self.verbose else arrays
        spectra = [self.fourier_transform(array) for array in arrays]
        self.sort_spectra(spectra, times)
        result_spectrum = np.empty_like(spectra[0])
        last_index = result_spectrum.shape[2]
        indices = (last_index * times).astype(int).tolist()
        indices.append(last_index)
        ini = 0
        for spectrum, fin in zip(spectra, indices):
            result_spectrum[..., ini:fin] = spectrum[..., ini:fin]
            ini = fin
        result_image = self.inv_fourier_transform(result_spectrum)
        return result_image.astype(np.float32)

    @staticmethod
    def fourier_transform(array):
        transformed = np.fft.fft2(array)
        fshift = np.fft.fftshift(transformed)
        return fshift

    @staticmethod
    def inv_fourier_transform(fshift):
        f_ishift = np.fft.ifftshift(fshift)
        img_back = np.fft.ifft2(f_ishift)
        return np.abs(img_back)

    def matrix_average(self, matrices, weights=None):
        if weights is None:
            num_matrices = len(matrices)
            weights = num_matrices * (1 / num_matrices,)
        logs = [w * logm(A) for (w, A) in zip(weights, matrices)]
        logs = np.array(logs)
        logs_sum = logs.sum(axis=0)
        return expm(logs_sum)


def get_params_array(nums_range, num_transforms):
    tensor = torch.FloatTensor(num_transforms, 3).uniform_(*nums_range)
    return tensor.numpy()



import torch
from ..torchio import DATA
from ..utils import is_image_dict
from .transform import Transform


class Lambda(Transform):
    def __init__(self, function, types_to_apply=None, verbose=False):
        super().__init__(verbose=verbose)
        self.function = function
        self.types_to_apply = types_to_apply

    def apply_transform(self, sample):
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            image_type = image_dict['type']
            has_types_list = self.types_to_apply is not None
            if has_types_list and image_type in self.types_to_apply:
                function_arg = image_dict[DATA][0]
                result = self.function(function_arg)
                if not isinstance(result, torch.Tensor):
                    message = (
                        'The returned value from the callable argument must be'
                        f' of type {torch.Tensor}, not {type(result)}'
                    )
                    raise ValueError(message)
                if result.dtype != torch.float32:
                    message = (
                        'The data type of the returned value must be'
                        f' of type {torch.float32}, not {result.dtype}'
                    )
                    raise ValueError(message)
                if result.ndim != function_arg.ndim:
                    message = (
                        'The number of dimensions of the returned value must'
                        f' be {function_arg.ndim}, not {result.ndim}'
                    )
                    raise ValueError(message)
                image_dict[DATA][0] = result
        return sample



import copy
from itertools import cycle

import numpy as np
import torch
from torch.utils.data import IterableDataset

from ...torchio import DATA
from ...utils import to_tuple, is_image_dict, check_consistent_shape


class ImageSampler(IterableDataset):
    def __init__(self, sample, patch_size):
        self.sample = sample
        self.patch_size = np.array(to_tuple(patch_size, n=3), dtype=np.uint16)

    def __iter__(self):
        return self.get_stream(self.sample, self.patch_size)

    def get_stream(self, sample, patch_size):
        """
        Is cycle neccesary?
        """
        return cycle(self.extract_patch_generator(sample, patch_size))

    def extract_patch_generator(self, sample, patch_size):
        while True:
            yield self.extract_patch(sample, patch_size)

    def extract_patch(self, sample, patch_size):
        index_ini, index_fin = self.get_random_indices(sample, patch_size)
        cropped_sample = self.copy_and_crop(
            sample,
            index_ini,
            index_fin,
        )
        return cropped_sample

    def get_random_indices(self, sample, patch_size):
        check_consistent_shape(sample)
        first_image_name = list(sample.keys())[0]
        first_image_array = sample[first_image_name][DATA]
        # first_image_array should have shape (1, H, W, D)
        shape = np.array(first_image_array.shape[1:], dtype=np.uint16)
        max_index_ini = shape - patch_size
        if (max_index_ini < 0).any():
            message = (
                f'Patch size {patch_size} must not be'
                f' larger than image size {tuple(shape)}'
            )
            raise ValueError(message)
        coordinates = []
        for max_coordinate in max_index_ini.tolist():
            if max_coordinate == 0:
                coordinate = 0
            else:
                coordinate = torch.randint(max_coordinate, size=(1,)).item()
            coordinates.append(coordinate)
        index_ini = np.array(coordinates, np.uint16)
        index_fin = index_ini + patch_size
        return index_ini, index_fin

    @staticmethod
    def crop(image, index_ini, index_fin):
        i_ini, j_ini, k_ini = index_ini
        i_fin, j_fin, k_fin = index_fin
        return image[..., i_ini:i_fin, j_ini:j_fin, k_ini:k_fin]

    def copy_and_crop(self, sample, index_ini, index_fin):
        cropped_sample = {}
        for key, value in sample.items():
            cropped_sample[key] = copy.copy(value)
            if is_image_dict(value):
                sample_image_dict = value
                cropped_image_dict = cropped_sample[key]
                cropped_image_dict[DATA] = self.crop(
                    sample_image_dict[DATA], index_ini, index_fin)
        # torch doesn't like uint16
        cropped_sample['index_ini'] = index_ini.astype(int)
        return cropped_sample



import warnings
from pathlib import Path
from collections.abc import Sequence
import torch
from torch.utils.data import Dataset
from ..utils import get_stem
from ..torchio import DATA, AFFINE
from .io import read_image, write_image


class ImagesDataset(Dataset):
    def __init__(
            self,
            subjects_list,
            transform=None,
            check_nans=True,
            verbose=False,
            ):
        """
        Each element of subjects_list should be an instance of torchio.Subject
        """
        self.parse_subjects_list(subjects_list)
        self.subjects_list = subjects_list
        self._transform = transform
        self.check_nans = check_nans
        self.verbose = verbose

    def __len__(self):
        return len(self.subjects_list)

    def __getitem__(self, index):
        subject_images = self.subjects_list[index]
        sample = {}
        for image in subject_images:
            tensor, affine = image.load(check_nans=self.check_nans)
            image_dict = {
                DATA: tensor,
                AFFINE: affine,
                'type': image.type,
                'path': str(image.path),
                'stem': get_stem(image.path),
            }
            sample[image.name] = image_dict

        # Apply transform (this is usually the bottleneck)
        if self._transform is not None:
            sample = self._transform(sample)
        return sample

    def set_transform(self, transform):
        self._transform = transform

    @staticmethod
    def parse_subjects_list(subjects_list):
        # Check that it's list or tuple
        if not isinstance(subjects_list, Sequence):
            raise TypeError(
                f'Subject list must be a sequence, not {type(subjects_list)}')

        # Check that it's not empty
        if not subjects_list:
            raise ValueError('Subjects list is empty')

        # Check each element
        for subject_list in subjects_list:
            subject = Subject(*subject_list)

    @classmethod
    def save_sample(cls, sample, output_paths_dict):
        for key, output_path in output_paths_dict.items():
            tensor = sample[key][DATA][0]  # remove channels dim
            affine = sample[key][AFFINE]
            write_image(tensor, affine, output_path)


class Subject(list):
    def __init__(self, *images, name=None):
        self.parse_images(images)
        super().__init__(images)
        self.name = name

    @staticmethod
    def parse_images(images):
        # Check that each element is a list
        if not isinstance(images, Sequence):
            message = (
                'Subject "images" parameter must be a sequence'
                f', not {type(images)}'
            )
            raise TypeError(message)

        # Check that it's not empty
        if not images:
            raise ValueError('Images list is empty')

        # Check that there are only instances of Image
        # and all images have different names
        names = []
        for image in images:
            if not isinstance(image, Image):
                message = (
                    'Subject list elements must be instances of'
                    f' torchio.Image, not {type(image)}'
                )
                raise TypeError(message)
            if image.name in names:
                message = (
                    f'More than one image with name "{image.name}"'
                    ' found in images list'
                )
                raise KeyError(message)
            names.append(image.name)


class Image:
    def __init__(self, name, path, type_):
        self.name = name
        self.path = self.parse_path(path)
        self.type = type_

    def parse_path(self, path):
        try:
            path = Path(path).expanduser()
        except TypeError:
            message = f'Conversion to path not possible for variable: {path}'
            raise TypeError(message)
        if not path.is_file():
            message = (
                f'File for image "{self.name}"'
                f' not found: "{path}"'
                )
            raise FileNotFoundError(message)
        return path

    def load(self, check_nans=True):
        tensor, affine = read_image(self.path)
        tensor = tensor.unsqueeze(0)  # add channels dimension
        if check_nans and torch.isnan(tensor).any():
            warnings.warn(f'NaNs found in file "{self.path}"')
        return tensor, affine




import torch
import SimpleITK as sitk
from ....utils import is_image_dict, nib_to_sitk, sitk_to_nib
from ....torchio import DATA, AFFINE, INTENSITY
from .. import RandomTransform


class RandomBlur(RandomTransform):
    def __init__(self, std_range=(0, 4), seed=None, verbose=False):
        super().__init__(seed=seed, verbose=verbose)
        self.std_range = std_range

    def apply_transform(self, sample):
        std = self.get_params(self.std_range)
        sample['random_blur'] = std
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] != INTENSITY:
                continue
            image_dict[DATA][0] = blur(
                image_dict[DATA][0],
                image_dict[AFFINE],
                std,
            )
        return sample

    @staticmethod
    def get_params(std_range):
        std = torch.FloatTensor(3).uniform_(*std_range).numpy()
        return std


def blur(data, affine, std):
    image = nib_to_sitk(data, affine)
    image = sitk.DiscreteGaussian(image, std.tolist())
    array, _ = sitk_to_nib(image)
    tensor = torch.from_numpy(array)
    return tensor



from pathlib import Path
import nrrd
import torch
import numpy as np
import nibabel as nib


def parse_path(path):
    path = Path(path).expanduser()
    suffixes = path.suffixes
    if '.nii' not in suffixes and '.nrrd' not in suffixes:
        raise ValueError(
            'Image suffixes must contain ".nii" or ".nrrd",'
            f' but they are {suffixes}')
    return path


def read_image(path):
    path = parse_path(path)
    suffixes = path.suffixes
    if '.nii' in suffixes:
        read = _read_nifti
    elif '.nrrd' in suffixes:
        read = _read_nrrd
    else:
        raise NotImplementedError(
            f'Reading not implemented for this format: "{path}"')
    return read(path)


def _read_nifti(path):
    nii = nib.load(str(path), mmap=False)
    ndims = len(nii.shape)
    assert ndims == 3
    data = nii.get_fdata(dtype=np.float32)
    tensor = torch.from_numpy(data)
    affine = nii.affine
    return tensor, affine


def _read_nrrd(path):
    data, header = nrrd.read(path)
    data = data.astype(np.float32)
    tensor = torch.from_numpy(data)
    affine = np.eye(4)
    affine[:3, :3] = header['space directions'].T
    affine[:3, 3] = header['space origin']
    lps_to_ras = np.diag((-1, -1, 1, 1))
    affine = np.dot(lps_to_ras, affine)
    return tensor, affine


def write_image(tensor, affine, path):
    path = Path(path)
    suffixes = path.suffixes
    if '.nii' in suffixes:
        write = _write_nifti
    elif '.nrrd' in suffixes:
        write = _write_nrrd
    else:
        raise NotImplementedError(
            f'Writing not implemented for this format: "{path}"')
    write(tensor, affine, path)


def _write_nifti(tensor, affine, path):
    nii = nib.Nifti1Image(tensor.numpy(), affine)
    nii.header['qform_code'] = 1
    nii.header['sform_code'] = 0
    nii.to_filename(str(path))


def _write_nrrd(tensor, affine, path):
    raise NotImplementedError




import torch
from ....utils import is_image_dict
from ....torchio import DATA, INTENSITY
from .. import RandomTransform


class RandomNoise(RandomTransform):
    def __init__(self, std_range=(0, 0.25), seed=None, verbose=False):
        super().__init__(seed=seed, verbose=verbose)
        self.std_range = std_range

    def apply_transform(self, sample):
        std = self.get_params(self.std_range)
        sample['random_noise'] = std
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            if image_dict['type'] != INTENSITY:
                continue
            image_dict[DATA] = add_noise(image_dict[DATA], std)
        return sample

    @staticmethod
    def get_params(std_range):
        std = torch.FloatTensor(1).uniform_(*std_range).item()
        return std


def add_noise(data, std):
    noise = torch.FloatTensor(*data.shape).normal_(mean=0, std=std)
    data = data + noise
    return data



from .transform import Transform
from .interpolation import Interpolation

# Generic
from .lambda_transform import Lambda

# Augmentation
from .augmentation.spatial import RandomFlip
from .augmentation.spatial import RandomAffine
from .augmentation.spatial import RandomElasticDeformation

from .augmentation.intensity import RandomBlur
from .augmentation.intensity import RandomNoise
from .augmentation.intensity import RandomMotion
from .augmentation.intensity import RandomBiasField

# Preprocessing
from .preprocessing import Pad
from .preprocessing import Crop
from .preprocessing import Rescale
from .preprocessing import Resample
from .preprocessing import ToCanonical
from .preprocessing import ZNormalization
from .preprocessing import CenterCropOrPad
from .preprocessing import HistogramStandardization



import numpy as np
import SimpleITK as sitk
from .pad import Pad
from .crop import Crop
from .bounds_transform import BoundsTransform
from ....torchio import DATA
from ....utils import is_image_dict, check_consistent_shape


class CenterCropOrPad(BoundsTransform):
    def __init__(
            self,
            size,
            padding_mode=None,
            padding_fill=None,
            verbose=False,
            ):
        """
        size should be an integer or a tuple of 3 integers
        """
        super().__init__(size, verbose=verbose)
        self.padding_mode = padding_mode
        self.padding_fill = padding_fill

    def apply_transform(self, sample):
        source_shape = self.get_sample_shape(sample)
        target_shape = np.array(self.bounds_parameters[::2])  # hack
        diff_shape = target_shape - source_shape

        cropping = -np.minimum(diff_shape, 0)
        if cropping.any():
            cropping_params = self.get_six_bounds_parameters(cropping)
            sample = Crop(cropping_params)(sample)

        padding = np.maximum(diff_shape, 0)
        if padding.any():
            padding_kwargs = {'fill': self.padding_fill}
            if self.padding_mode is not None:
                padding_kwargs['padding_mode'] = self.padding_mode
            padding_params = self.get_six_bounds_parameters(padding)
            sample = Pad(padding_params, **padding_kwargs)(sample)
        return sample

    @staticmethod
    def get_sample_shape(sample):
        check_consistent_shape(sample)
        for image_dict in sample.values():
            if not is_image_dict(image_dict):
                continue
            return image_dict[DATA].shape[1:]  # remove channels dimension

    @staticmethod
    def get_six_bounds_parameters(parameters):
        parameters = parameters / 2
        result = []
        for n in parameters:
            ini, fin = int(np.ceil(n)), int(np.floor(n))
            result.extend([ini, fin])
        return tuple(result)
